local MAJOR = "LibScriptableDisplayBuffer-1.0"
local MINOR = 15
assert(LibStub, MAJOR.." requires LibStub")
local LibBuffer = LibStub:NewLibrary(MAJOR, MINOR)
if not LibBuffer then return end
local LibError = LibStub("LibScriptableDisplayError-1.0")
assert(LibError, MAJOR .. " requires LibScriptableDisplayError-1.0")

local pool = setmetatable({}, {__mode = "k"})

if not LibBuffer.__index then
	LibBuffer.__index = LibBuffer
end

local new, del
do
	local pool = setmetatable({}, {__mode = "k"})
	function new(size)
		local obj = next(pool)
		if obj then
			local smallest = obj
			for tbl in pairs(pool) do
				if #tbl >= size and #tbl < #smallest then
					smallest = tbl
				end
			end
			obj = smallest
			pool[obj] = nil
		else
			obj = setmetatable({}, {__index=function(this, i)
				if type(i) ~= "number" then return nil end
				if rawget(this, i) == true then return nil end
				return rawget(this, i)
			end})
		end
		return obj
	end
	function del(obj)
		pool[obj] = true
	end
end

-- @name LibScriptableDisplayBuffer:New
-- @usage LibScriptableDisplayBuffer:New(name, size, defval, errorLevel)
-- @param name A name for the buffer
-- @param size The buffer's size
-- @param defval The buffer's default value. The alpha channel.
-- @param errorLevel The verbocity level
-- @return A new LibScriptableDisplayBuffer object

function LibBuffer:New(name, size, defval, errorLevel)
	assert(type(name) ~= "nil", "LibBuffer requires a name")
	assert(type(size) ~= "nil", "Buffer size is required")
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	obj.buffer = new(size)

	setmetatable(obj, self)

	obj.name = name
	obj.defval = defval or " "
	obj.error = LibError:New(MAJOR .. " " .. name, errorLevel)
	obj:Resize(size)

	return obj

end

-- @name LibScriptableDisplayBuffer:Del
-- @usage LibScriptableDisplayBuffer:Del([lcd]) or object:Del()
-- @param lcd An optional lcd object
-- @return Nothing
function LibBuffer:Del(buff)
	if not buff then buff = self end
	pool[buff] = true
	buff.error:Del()
	for i = 0, buff:Size() do
		buff.buffer[i] = false
	end
	buff.size = 0
	del(buff.buffer)
end

-- @name :Wipe
-- @usage :Wipe()
-- @return Nothing
function LibBuffer:Wipe()
	for i = 0, self:Size() - 1 do
		self.buffer[i] = self.defval
	end
end

-- @name :Fill
-- @usage :Fill(val)
-- @param val The value to fill the buffer with
-- @return Nothing
function LibBuffer:Fill(val)
	for i = 0, self:Size() - 1 do
		self.buffer[i] = val
	end
end

-- @name :Resize
-- @usage :Resize(size)
-- @param size The buffer's new size
-- @return Nothing
function LibBuffer:Resize(size)
	assert(self.defval ~= nil)
	assert(type(size) == "number")
	if self:Size() < size then
		for i = self:Size(), size - 1 do
			self.buffer[i] = self.defval
		end
	end

	if self:Size() > size then
		for i = size - 1, self:Size() do
			self.buffer[i] = false
		end
	end

	self.size = size

	if size > 0 then
		for i = 0, size - 1 do
			assert(self.buffer[i] ~= nil, format("resize %d %d", i, size))
		end
	end
end

function LibBuffer:Replace(pos, val)
	assert(type(pos) == "number", "Position sent to Replace isn't a number")
	assert(val ~= nil, "Val sent to Replace is nil")
	if pos > self.size then self.size = pos end
	self.buffer[pos] = val
end

function LibBuffer:Insert(i, val)
	local sub = self:Sub(i + 1)
	self.buffer[0] = val
	self:Merge(sub, 1, sub:Size() - 1)
	sub:Del()
end

function LibBuffer:Clear()
	for i = 0, self:Size() - 1 do
		self.buffer[i] = self.defval
	end
end

local pool = setmetatable({}, {__mode="k"})
local function new()
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	wipe(obj)
	return obj
end

local function del(obj)
	pool[obj] = true
end

function LibBuffer:AsString()
	assert(type(self.defval) == "string")
	local tbl = new()
	for i = 0, self:Size() - 1 do
		if type(self.buffer[i]) == "string" then
			tbl[i + 1] = self.buffer[i]
		else
			tbl[i + 1] = ""
		end
	end
	local str = table.concat(tbl)
	del(tbl)

	--local str = (self.buffer[0] or "") .. table.concat(self.buffer)
	--[[
	for i = 0, self:Size() - 1 do
		if self.buffer[i] == nil then break end
		if type(self.buffer[i]) == "string" then
			str = str .. self.buffer[i]
		else
			str = str .. self.defval
			self.error:Print(":AsString() - Bad index: " .. i .. ", type: " .. type(self.buffer[i]) .. ", size: " .. self:Size())
		end
	end]]
	return str
end

function LibBuffer:FromString(str)
	if type(str) ~= "string" then return end
	for i = 0, strlen(str) - 1 do
		self.buffer[i] = str:sub(i + 1, i + 1)
	end
	if strlen(str) < self:Size() then
		for i = strlen(str) - 1, self:Size() do
			self.buffer[i] = false
		end
	end
	self.size = strlen(str)
end

function LibBuffer:MergeString(pos1, str, pos2, length)
	for n = 0, length - 1 do
		self.buffer[pos1 + n] = str:sub(pos2, pos2 + length)
	end
end

function LibBuffer:FromList(list)
	if type(list) ~= "table" then return end
	for i, v in ipairs(list) do
		self.buffer[i - 1] = v
	end
	self.size = #list
end

function LibBuffer:Size()
	--[[
	local count = 0
	for k in pairs(self.buffer) do
		count = count + 1
	end
	return count
	]]
	return self.size or 0
end

function LibBuffer:Sub(num1, num2)

	if not num2 then num2 = self:Size() - 1 end
	if num1 > num2 or num2 > self:Size() - 1 then return end
	
	local t = LibBuffer:New(self.name .. ":" , num2 - num1 + 1, self.defval, self.errorLevel)
	for i = num1, num2 do
		t.buffer[i - num1] = self.buffer[i] or self.defval
	end
	return t
end

function LibBuffer:Copy(other)
	table.wipe(self.buffer)
	for k, v in pairs(other.buffer) do
		self.buffer[k] = other.buffer[k]
	end
end

function LibBuffer:Merge(from, pos, len)
	local i = 0
	while i < len - (self.size - len) do
		assert(from.buffer[i] ~= nil)
		self.buffer[i + pos] = from.buffer[i]
		i = i + 1
	end
end

function LibBuffer:Memset(pos, char, length)
	for n = pos, pos + length - 1 do
		self.buffer[n] = char
	end
end

function LibBuffer:Memcopy(pos1, buffer, pos2, length)
	for n = 0, length - 1 do
		self.buffer[pos1 + n] = buffer.buffer[pos2 + n] or self.defval
	end
end

local function randomize(a, b)
	if random(2) == 1 then
		return a < b
	else
		return a > b
	end
end

local function copy(tbl)
	if type(tbl) ~= "table" then return tbl end
	local new = new(#tbl)
	for k, v in pairs(tbl) do
		new[k + 1] = copy(v)
	end
	return new
end

function LibBuffer:Randomize()
	local c = copy(self.buffer)
	local len = #c
	for i = 0, self:Size() - 1 do
		local i = random(#c)
		self.buffer[i] = c[i]
		tremove(c, i)
	end
	for i = 0, len - 1 do
		c[i] = true
	end
	del(c)	
	return self
end

function LibBuffer:MovingAverageExp(alpha, epsilon, buf)
	if type(epsilon) ~= "number" then epsilon = 0 end

	if not (0 < alpha and alpha < 1) then
		return
	end

	if not (0 <= epsilon and epsilon < alpha) then
		return
	end
	
	result = buf or LibBuffer:New("result", self:Size(), self.errorLevel)
	
	for i = 0, result:Size() - 1 do
		local currentWeight = 1.0
		local numerator = 0
		local denominator = 0
		local sub = self:Sub(i)
		for j = 0, sub:Size() - 1 do
			local value = sub.buffer[j]
			numerator = numerator + value * currentWeight
			denominator = denominator + currentWeight

			currentWeight = currentWeight * alpha
			if currentWeight < epsilon then break end
		end
		result:Replace(i, numerator / denominator)
		sub:Del()
	end

	return result
end

function LibBuffer:Line(x0, y0, x1, y1, color, pitch)
	local dx = x1 - x0;
	local dy = y1 - y0;

	self.buffer[x0 + y0 * pitch] = color

	if (dx ~= 0) then
		local m = dy / dx;
		local b = y0 - m*x0;
		if x1 > x0 then
			dx = 1
		else
			dx = -1
		end
		while x0 ~= x1 do
			x0 = x0 + dx
			y0 = math.floor(m*x0 + b + 0.5);
			self.buffer[x0 + y0 * pitch] = color
		end
        
    end
end

function LibBuffer:Line2(x0, y0, x1, y1, color, pitch)
	local dx = x1 - x0;
	local dy = y1 - y0;

	self.buffer[x0 + y0 * pitch] = color
	if math.abs(dx) > math.abs(dy) then -- slope < 1
		local m = dy / dx -- compute slope
		local b = y0 - m*x0
		if dx < 0 then
			dx = -1
		else
			dx = 1
		end
		while x0 ~= x1 do
			x0 = x0 + dx
			local n = floor(m*x0 + b + 0.5) * pitch + x0
			self.buffer[n] = color
		end
		
	elseif dy ~= 0 then -- slope >= 1
		local m = dx / dy -- compute slope
		local b = x0 - m*y0
		if dx < 0 then
			dx = -1
		else
			dx = 1
		end
		while y0 ~= y1 do
			y0 = y0 + dy
			local n = y0 * pitch + floor(m*y0 + b + 0.5)
			self.buffer[n] = color
		end
	end
end

function LibBuffer:Line3(x0, y0, x1, y1, color, pitch)
	local dx = x1 - x0;
	local dy = y1 - y0;
	local t = 0.5

	self.buffer[x0 + y0 * pitch] = color
	if math.abs(dx) > math.abs(dy) then -- slope < 1
		local m = dy / dx -- compute slope
		t = t + y0
		--local b = y0 - m*x0
		if dx < 0 then
			dx = -1
		else
			dx = 1
		end
		m = m * dx
		while x0 ~= x1 do
			x0 = x0 + dx
			t = t + m
			self.buffer[ t * pitch + x0] = color
		end
		
	elseif dy ~= 0 then -- slope >= 1
		local m = dx / dy -- compute slope
		t = t + x0
		if dy < 0 then
			dy = -1
		else
			dy = 1
		end
		m = m * dy
		while y0 ~= y1 do
			y0 = y0 + dy
			t = t + m
			self.buffer[y0 * pitch + x0] = color
		end
	end
end

function LibBuffer:Line4(x0, y0, x1, y1, color, pitch)
	local dx = x1 - x0;
	local dy = y1 - y0;
	local stepx, stepy
	
	if dy < 0 then
		dy = -dy
		stepy = -1
	else
		stepy = 1
	end
	
	if dx < 0 then
		dx = -dx
		stepx = -1
	else
		stepx = 1
	end

	self.buffer[x0 + y0 * pitch] = color
	if dx > dy then
		local fraction = dy - bit.rshift(dx, 1)
		while x0 ~= x1 do
			if fraction >= 0 then
				y0 = y0 + stepy
				fraction = fraction - dx
			end
			x0 = x0 + stepx
			fraction = fraction + dy
			self.buffer[floor(y0) * pitch + floor(x0)] = color
		end
	else
		local fraction = dx - bit.rshift(dy, 1)
		while y0 ~= y1 do
			if fraction >= 0 then
				x0 = x0 + stepx
				fraction = fraction - dy
			end
			y0 = y0 + stepy
			fraction = fraction + dx
			self.buffer[floor(y0) * pitch + floor(x0)] = color
		end
	end
end
