local MAJOR = "LibScriptableDisplayWidgetColor-1.0"
local MINOR = 15

assert(LibStub, MAJOR.." requires LibStub")
local WidgetColor = LibStub:NewLibrary(MAJOR, MINOR)
if not WidgetColor then return end
local LibWidget = LibStub("LibScriptableDisplayWidget-1.0", true)
assert(LibWidget, MAJOR .. " requires LibScriptableDisplayWidget-1.0")
local LibTimer = LibStub("LibScriptableDisplayTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires LibScriptableDisplayTimer-1.0")
local Evaluator = LibStub("LibScriptableDisplayEvaluator-1.0", true)
assert(Evaluator, MAJOR .. " requires LibScriptableDisplayEvaluator-1.0")
local PluginUtils = LibStub("LibScriptableDisplayPluginUtils-1.0", true)
assert(PluginUtils, MAJOR .. " requires LibScriptableDisplayPluginUtils-1.0")
local L = LibStub("LibScriptableDisplayLocale-1.0", true)
assert(L, MAJOR .. " requires LibScriptableDisplayLocale-1.0")
L = L.L

if not WidgetColor.__index then WidgetColor.__index = WidgetColor end


local defaults = {
	update = 500,
	repeating = true
}
WidgetColor.defaults = defaults


local widgetType = {gesture=true}

local pool = setmetatable({}, {__mode = "v"})
local function new(...)
	local obj = table.remove(pool) or {}
	for i = 1, select("#", ...) do
		obj[i] = select(i, ...)
	end
	return obj
end

local function newDict(...)
	local obj = table.remove(pool) or {}
	for i = 1, select("#", ...), 2 do
		local key = select(i, ...)
		local val = select(i + 1, ...)
		obj[key] = val
	end
	return obj
end

local function del(obj)
	wipe(obj)
	tinsert(pool, obj)
end

local pool = setmetatable({}, {__mode = "k"})

local stopFunc

-- Create a new LibScriptableDisplayWidgetColor object
-- @name WidgetColor.New
-- @usage WidgetColor:New(visitor, name, config, errorLevel)
-- @param visitor An LibScriptableDisplayCore-1.0 object, or provide your own
-- @param name A name for the timer widget
-- @param config This timeer's parameters
-- @param errorLevel The errorLevel for this object
-- @param callback An optional callback function to be executed when the gesture is performed.
-- @param An optional timer. This should have a :Start() and :Stop().
-- @return A new LibScriptableDisplayWidgetColor widget
function WidgetColor:New(visitor, name, config, errorLevel, callback, timer)
	assert(name, "WidgetColor requires a name.")
	assert(config, "Please provide the timer with a config")
	assert(config.expression, name .. ": Please provide the timer with an expression")

	local obj = next(pool)

	if obj then
		pool[obj] = nil
		obj.__index = nil
	else
		obj = {}
		obj.options = {}
	end

	setmetatable(obj, self)

	obj.widget = LibWidget:New(obj, visitor, MAJOR .. " " .. name, config, 0, 0, 0, widgetType, errorLevel)

	obj.expression = config.expression

	obj.update = config.update or defaults.update
	
	obj.draw = callback
	
	obj.repeating = config.repeating or defaults.repeating

	obj.timer = timer or LibTimer:New("WidgetColor.timer " .. obj.name, obj.update, obj.repeating, self.Update, obj, obj.errorLevel)

	return obj
end


-- Delete a LibScriptableDisplayWidgetColor object
-- @name LibScriptableDisplayWidgetColor.Del
-- @usage LibScriptableDisplayWidgetColor:Del([marq])
-- @param marq A LibScriptableDisplayWidgetColor object
-- @return Nothing
function WidgetColor:Del()
	self:Stop()
	self.widget:Del()
	self.timer:Del()
	pool[self] = true
end

-- Start a LibScriptableDisplayWidgetColor
-- @name LibScriptableDisplayWidgetColor.Start
-- @usage object:Start()
-- @return Nothing
function WidgetColor:Start()
	if self.update > 0 then
		self.timer:Start()
		self.active = true
	end
	self:Update()
end

-- Stop a LibScriptableDisplayWidgetColor
-- @name LibScriptableDisplayWidgetColor.Stop
-- @usage object:Stop()
-- @return Nothing
function WidgetColor:Stop()
	self.timer:Stop()
	self.active = false
end

function WidgetColor:Update()

	self.environment.self = self
	local r, g, b, a = Evaluator.ExecuteCode(self.environment, MAJOR .. " Lua script", self.expression)
	--self.environment.self = nil

	if type(r) ~= "number" then return end

	if r and r > 1 then
		r = r / 255
	end

	if g and g > 1 then
		g = g / 255
	end

	if b and b > 1 then
		b = b / 255
	end

	if a and a > 1 then
		a = a / 255
	end

	if not g then
		if r > bit.bor(bit.bor(0xff, 0xff), 0xff) then
			self.r, self.g, self.b, self.a = PluginUtils.Color2RGBA(r, true)
		end
	else
		self.r, self.g, self.b, self.a = r, g, b, a
	end
	
	if type(self.draw) == "function" then
		self.draw(self)
	end
end

-- Get an Ace3 option table. Plug this into a group type's args.
-- @param db The database table
-- @param callback Provide this if you want to execute the callback once an option is changed
-- @param data Some data to pass when executing the callback
-- @return An Ace3 options table -- `name.args = options`.
function WidgetColor:GetOptions(db, callback, data)
	local defaults = WidgetColor.defaults
	local options = {
		enabled = {
			name = "Enabled",
			desc = "Whether this timer is enabled or not",
			type = "toggle",
			get = function() return db.enabled end,
			set = function(info, v)
				db.enabled = v
				db.enabledDirty = true
				if type(callback) == "function" then callback(data) end
			end,
			order = 5
		},
		update = {
			name = "Update Rate",
			desc = "Enter the timer's refresh rate",
			type = "input",
			pattern = "%d",
			get = function()
				return tostring(db.update or defaults.update)
			end,
			set = function(info, v)
				db.update = tonumber(v)
				db.updateDirty = true
				if type(callback) == "function" then callback(data) end
			end,
			order = 6
		},
		repeating = {
			name = "Repeating Timer",
			desc = "Whether the timer associated with this widget repeats or not",
			type = "toggle",
			get = function()
				return db.repeating
			end,
			set = function(info, val)
				db.repeating = val
				db.repeatingDirty = true
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 7
		},
		expression = {
			name = "Expression",
			desc = "This widget's Lua script",
			type = "input",
			multiline = true,
			width = "full",
			get = function()
				return db.expression
			end,
			set = function(info, val)
				db.expression = val
				db.expressionDirty = val
				if type(callback) == "function" then
					callback(data)
				end
			end,
			order = 7
		}
	}
	return options
end
